unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls;

type
  TSeverity = (seOK, seWarning, seCritical);

  TForm1 = class(TForm)
    Button1: TButton;
    Memo1: TMemo;
    cbVerbose: TCheckBox;
    LabeledEdit1: TLabeledEdit;
    cbWarnMD5Missing: TCheckBox;
    cbWarningMissingMD5Entry: TCheckBox;
    cbWarnVanishedFile: TCheckBox;
    cbWarnChecksumMismatch: TCheckBox;
    Label1: TLabel;
    procedure Button1Click(Sender: TObject);
  private
    MD5filecount: integer;
    function CheckDirectory(ADirectory: string; recursive: boolean): TSeverity;
    function VerifyMD5File(amd5file: string): TSeverity;
    function md5filesafe(const filename: string): string;
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  IdHashMessageDigest, idHash;

const
  DUMMY_MD5 = 'DUMMY.MD5';

function md5file(const filename: string): string;
var
  IdMD5: TIdHashMessageDigest5;
  FS: TFileStream;
begin
  IdMD5 := TIdHashMessageDigest5.Create;
  FS := TFileStream.Create(filename, fmOpenRead or fmShareDenyWrite);
  try
{$IFDEF UNICODE} // I actually do not know at which version of Delphi/Indy, this has changed.
    Result := IdMD5.HashStreamAsHex(FS);
{$ELSE}
    Result := IdMD5.AsHex(IdMD5.HashValue(FS));
{$ENDIF}
  finally
    FS.Free;
    IdMD5.Free;
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  sev: TSeverity;
begin
  Memo1.Clear;
  if not DirectoryExists(LabeledEdit1.Text) then
  begin
    showmessage('Directory does not exist');
    exit;
  end;
  Application.ProcessMessages;
  MD5filecount := 0;
  sev := CheckDirectory(LabeledEdit1.Text, true);
  case sev of
    seOK:
      showmessage('OK');
    seWarning:
      showmessage('Warning');
    seCritical:
      showmessage('Critical');
  end;
  Caption := Format('Done. Checked %d MD5 files.', [MD5filecount]);
end;

procedure MD5FileToStringList(amd5file: string; slOut: TStringList);
var
  slLines: TStringList;
  sLine: string;
  originalFilename: string;
  expectedChecksum: string;
  i: integer;
begin
  if not FileExists(amd5file) then
    exit;

  slLines := TStringList.Create;
  try
    slLines.LoadFromFile(amd5file);
    for i := 0 to slLines.Count - 1 do
    begin
      sLine := slLines.Strings[i];
      // 25bfdef2651071efdd08bb3404797384 *TSEXITSTAC.exe
      sLine := Trim(sLine);
      if sLine = '' then
        continue;
      expectedChecksum := Copy(sLine, 1, 32);
      Delete(sLine, 1, 32);
      sLine := Trim(sLine);
      if Copy(sLine, 1, 1) = '*' then
        Delete(sLine, 1, 1);
      sLine := Trim(sLine);
      originalFilename := sLine;

      slOut.Values[originalFilename] := expectedChecksum;
    end;
  finally
    slLines.Free;
  end;
end;

function SevMax(a, b: TSeverity): TSeverity;
begin
  if Ord(a) > Ord(b) then
    Result := a
  else
    Result := b;
end;

function TForm1.VerifyMD5File(amd5file: string): TSeverity;
var
  slFile: TStringList;
  i: integer;
  originalFilename: string;
  expectedChecksum: string;
  IsFound: boolean;
  SR: TSearchRec;
  fullfilename: string;
  ADirectory: string;
  originalFilenameFull: string;
begin
  if cbVerbose.Checked then
  begin
    if ExtractFileName(amd5file) <> DUMMY_MD5 then
    begin
      Inc(MD5filecount);
    end;
    Form1.Memo1.Lines.Add('Check: ' + amd5file);
  end;

  Result := seOK;
  ADirectory := IncludeTrailingPathDelimiter(ExtractFilePath(amd5file));

  try
    slFile := TStringList.Create;
    try
      slFile.CaseSensitive := false;

      MD5FileToStringList(amd5file, slFile);
      // TODO: If multiple MD5 files => put them together into a single array (beware conflicts!)

      // 1. Check existing entries in the MD5 file

      for i := 0 to slFile.Count - 1 do
      begin
        originalFilename := slFile.KeyNames[i];
        expectedChecksum := slFile.Values[originalFilename];

        originalFilenameFull := ADirectory + originalFilename;

        if not FileExists(originalFilenameFull) then
        begin
          if cbWarnVanishedFile.Checked then
          begin
            Form1.Memo1.Lines.Add('FILE VANISHED: ' + originalFilenameFull);
            Result := SevMax(Result, seCritical);
          end;
        end
        else if LowerCase(md5filesafe(originalFilenameFull))
          = LowerCase(expectedChecksum) then
        begin
          if cbVerbose.Checked then
          begin
            Form1.Memo1.Lines.Add('OK: ' + originalFilenameFull + ' = ' +
              expectedChecksum);
          end;
          Result := SevMax(Result, seOK);
        end
        else
        begin
          if cbWarnChecksumMismatch.Checked then
          begin
            Form1.Memo1.Lines.Add('CHECKSUM MISMATCH: ' + originalFilenameFull +
              ' <> ' + expectedChecksum);
            Result := SevMax(Result, seCritical);
          end;
        end;
      end;

      // 2. Checking for entries which are NOT in the MD5 file

      IsFound := FindFirst(ADirectory + '*', faAnyFile - faDirectory, SR) = 0;
      while IsFound do
      begin
        fullfilename := ADirectory + SR.Name;
        if LowerCase(ExtractFileExt(fullfilename)) <> '.md5' then
        begin
          if slFile.Values[SR.Name] = '' then
          begin
            if ExtractFileName(amd5file) = DUMMY_MD5 then
            begin
              if cbWarnMD5Missing.Checked then
              begin
                Form1.Memo1.Lines.Add('NEW FILE WITHOUT CHECKSUM FILE: ' +
                  fullfilename);
                Result := SevMax(Result, seWarning);
              end;
            end
            else
            begin
              if cbWarningMissingMD5Entry.Checked then
              begin
                Form1.Memo1.Lines.Add('NEW FILE WITHOUT CHECKSUM ENTRY: ' +
                  fullfilename);
                Result := SevMax(Result, seWarning);
              end;
            end;
          end;
        end;
        IsFound := FindNext(SR) = 0;
      end;
      FindClose(SR);
    finally
      slFile.Free;
    end;
  except
    on E: Exception do
    begin
      Memo1.Lines.Add('Invalid MD5 file: ' + amd5file + ' : ' + E.Message);
      Result := seCritical;
    end;
  end;
end;

function TForm1.CheckDirectory(ADirectory: string; recursive: boolean)
  : TSeverity;
var
  IsFound: boolean;
  SR: TSearchRec;
  fullfilename: string;
begin
  Caption := ADirectory;
  Result := seOK;
  ADirectory := IncludeTrailingPathDelimiter(ADirectory);

  // Check MD5 files
  IsFound := FindFirst(ADirectory + '*.md5', faAnyFile - faDirectory, SR) = 0;
  if not IsFound then
  begin
    fullfilename := ADirectory + DUMMY_MD5; // virtual "empty" file
    Result := SevMax(Result, VerifyMD5File(fullfilename));
  end
  else
  begin
    while IsFound do
    begin
      fullfilename := ADirectory + SR.Name;
      Result := SevMax(Result, VerifyMD5File(fullfilename));
      IsFound := FindNext(SR) = 0;
    end;
  end;
  FindClose(SR);

  // Check other dirs
  if recursive then
  begin
    IsFound := FindFirst(ADirectory + '*', faAnyFile, SR) = 0;
    while IsFound do
    begin
      fullfilename := ADirectory + SR.Name;
      if DirectoryExists(fullfilename) and (SR.Name <> '.') and (SR.Name <> '..')
      then
      begin
        Result := SevMax(Result, CheckDirectory(fullfilename, recursive));
      end;
      IsFound := FindNext(SR) = 0;
    end;
    FindClose(SR);
  end;
end;

function TForm1.md5filesafe(const filename: string): string;
begin
  try
    Result := md5file(filename);
  except
    on E: Exception do
    begin
      Memo1.Lines.Add('Cannot read file ' + filename + ' : ' + E.Message);
      Result := 'ERROR';
    end;
  end;
end;

end.
